from .wrapper_opencv_window import WrapperOpenCVWindow


class StreamWindow(WrapperOpenCVWindow):
    """
    a basic window that displays the stream from a stream receiver
    
    :type self.wrap_object: gbvision.StreamReceiver
    """

    def _get_frame(self):
        return self.wrap_object.get_frame()



import abc
from threading import Thread

from gbvision.constants.types import Frame
from gbvision.exceptions.stream_closed import StreamClosed
from gbvision.net.stream_receiver import StreamReceiver


class AsyncStreamReceiver(StreamReceiver, abc.ABC):
    """
    an abstract async tcp stream receiver that receives frames on another thread
    None! when inheriting from this class and another StreamReceiver class, make sure you call the other class'
    constructor before this class' constructor, but also make sure you inherit from this class first in order
    """

    def __init__(self, *args, **kwargs):
        StreamReceiver.__init__(self, *args, **kwargs)
        self.__frame = None
        self.__thread = Thread(target=self.__receive_thread_function)
        self.__thread.start()

    def __receive_thread_function(self):
        while True:
            try:
                self.__frame = self._get_frame()
            except StreamClosed:
                break

    @abc.abstractmethod
    def _get_frame(self) -> Frame:
        """
        reads a frame from the stream synchronously (similar to StreamReceiver.get_frame).
        unsafe not to be used by programmer
        :return: the frame read
        """
        pass

    def get_frame(self):
        return self.__frame



from typing import List

import cv2
import numpy as np

from gbvision.constants.system import EMPTY_PIPELINE
from gbvision.constants.types import Frame, Polygon, Contour, Number, Point, FilterFunction
from gbvision.models.contours import FilterContours, find_contours, sort_polygons, contour_center, contours_to_polygons, \
    convex_hull_multiple
from .object_finder import ObjectFinder


class ConvexPolygonFinder(ObjectFinder):
    """
    finds any generic polygon, not recommended when another finder can be used

    :param area_scalar: optional, a scalar to multiply the area by, for fine tuning of the function's output
    :param contour_min_area: the minimal area of a contour, used for FilterContours, default is 0 (no area limit)
    :param threshold_func: a pipeline (or any sort of function) that returns a binary threshold of the object
     the finder is searching, the object needs to be white and the rest if the image black (doesn't
     have to be perfect)
    """

    def __init__(self, threshold_func: FilterFunction, game_object, area_scalar=1.0, contour_min_area=0.0):
        ObjectFinder.__init__(self, game_object, area_scalar=area_scalar)
        self._full_pipeline = (EMPTY_PIPELINE +
                               threshold_func +
                               find_contours +
                               convex_hull_multiple +
                               FilterContours(min_area=contour_min_area) +
                               contours_to_polygons +
                               sort_polygons)

    def find_shapes(self, frame: Frame) -> List[Polygon]:
        return self._full_pipeline(frame)

    @staticmethod
    def _shape_root_area(shape: Contour) -> Number:
        return np.sqrt(cv2.contourArea(shape))

    @staticmethod
    def _shape_center(shape: Contour) -> Point:
        return contour_center(shape)



from typing import List

from gbvision.constants.types import Frame, Rect, Number, Point, FilterFunction

from .object_finder import ObjectFinder
from gbvision.models.contours import find_contours, FilterContours, sort_contours, contours_to_rects_sorted
from gbvision.constants.system import EMPTY_PIPELINE
from gbvision.models.shapes import filter_inner_rects
import numpy as np


class RectFinder(ObjectFinder):
    """
    finds a rectangular shaped object

    :param area_scalar: optional, a scalar to multiply the area by, for fine tuning of the function's output
    :param contour_min_area: the minimal area of a contour, used for FilterContours, default is 0 (no area limit)
    :param threshold_func: a pipeline (or any sort of function) that returns a binary threshold of the object
     the finder is searching, the object needs to be white and the rest if the image black (doesn't
     have to be perfect)
    """

    def __init__(self, threshold_func: FilterFunction, game_object, area_scalar=1.0, contour_min_area=0.0):
        ObjectFinder.__init__(self, game_object, area_scalar=area_scalar)
        self._full_pipeline = (EMPTY_PIPELINE +
                               threshold_func +
                               find_contours +
                               FilterContours(min_area=contour_min_area) +
                               sort_contours +
                               contours_to_rects_sorted +
                               filter_inner_rects)

    def find_shapes(self, frame: Frame) -> List[Rect]:
        return self._full_pipeline(frame)

    @staticmethod
    def _shape_root_area(shape: Rect) -> Number:
        return np.sqrt(shape[2] * shape[3])

    @staticmethod
    def _shape_center(shape: Rect) -> Point:
        return (shape[0] + shape[2]) / 2, (shape[1] + shape[3]) / 2



from .recording_wrapper_opencv_window import RecordingWrapperOpenCVWindow
from .stream_window import StreamWindow


class RecordingStreamWindow(RecordingWrapperOpenCVWindow, StreamWindow):
    """
    a basic window that displays the stream from a stream receiver
    and records the stream to a video file
    """



from typing import List

from gbvision.constants.types import Circle, Frame, Number, Point, FilterFunction

from gbvision.constants.math import SQRT_PI
from gbvision.constants.system import EMPTY_PIPELINE
from gbvision.models.contours import find_contours, FilterContours, contours_to_circles_sorted
from gbvision.models.shapes import filter_inner_circles
from .object_finder import ObjectFinder


class CircleFinder(ObjectFinder):
    """
    finds specific circular shaped object, and performs distance transformation

    :param threshold_func: a pipeline (or any sort of function) that returns a binary threshold of the object
     the finder is searching, the object needs to be white and the rest if the image black (doesn't
     have to be perfect)
    :param contour_min_area: the minimal area of a contour, used for FilterContours, default is 0 (no area limit)
    """

    def __init__(self, threshold_func: FilterFunction, game_object, area_scalar=1.0, contour_min_area=0):
        ObjectFinder.__init__(self, game_object, area_scalar=area_scalar)
        self._full_pipeline = (EMPTY_PIPELINE +
                               threshold_func +
                               find_contours +
                               FilterContours(min_area=contour_min_area) +
                               contours_to_circles_sorted +
                               filter_inner_circles)

    def find_shapes(self, frame: Frame) -> List[Circle]:
        return self._full_pipeline(frame)

    @staticmethod
    def _shape_root_area(shape: Circle) -> Number:
        return SQRT_PI * shape[1]

    @staticmethod
    def _shape_center(shape: Circle) -> Point:
        return shape[0]
